# automatically generated by the FlatBuffers compiler, do not modify

# namespace: polygons

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Polygon:
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Polygon()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPolygon(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Polygon
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Polygon
    def Coords(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # Polygon
    def CoordsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Polygon
    def CoordsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Polygon
    def CoordsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def PolygonStart(builder):
    builder.StartObject(1)


def Start(builder):
    PolygonStart(builder)


def PolygonAddCoords(builder, coords):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(coords), 0
    )


def AddCoords(builder, coords):
    PolygonAddCoords(builder, coords)


def PolygonStartCoordsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartCoordsVector(builder, numElems):
    return PolygonStartCoordsVector(builder, numElems)


def PolygonEnd(builder):
    return builder.EndObject()


def End(builder):
    return PolygonEnd(builder)
