namespace timezonefinder.flatbuf.generated.polygons;

table Polygon {
  // combined x and y coordinates stored as a single vector
  coords:[int];
}

table PolygonCollection {
  polygons:[Polygon];
}

root_type PolygonCollection;
